﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  public abstract class TimeSeriesPrognosisSolutionBase : DataAnalysisSolution, ITimeSeriesPrognosisSolution {
    private const string TrainingMeanSquaredErrorResultName = "Mean squared error (training)";
    private const string TestMeanSquaredErrorResultName = "Mean squared error (test)";
    private const string TrainingMeanAbsoluteErrorResultName = "Mean absolute error (training)";
    private const string TestMeanAbsoluteErrorResultName = "Mean absolute error (test)";
    private const string TrainingSquaredCorrelationResultName = "Pearson's R² (training)";
    private const string TestSquaredCorrelationResultName = "Pearson's R² (test)";
    private const string TrainingRelativeErrorResultName = "Average relative error (training)";
    private const string TestRelativeErrorResultName = "Average relative error (test)";
    private const string TrainingNormalizedMeanSquaredErrorResultName = "Normalized mean squared error (training)";
    private const string TestNormalizedMeanSquaredErrorResultName = "Normalized mean squared error (test)";
    private const string TrainingDirectionalSymmetryResultName = "Average directional symmetry (training)";
    private const string TestDirectionalSymmetryResultName = "Average directional symmetry (test)";
    private const string TrainingWeightedDirectionalSymmetryResultName = "Average weighted directional symmetry (training)";
    private const string TestWeightedDirectionalSymmetryResultName = "Average weighted directional symmetry (test)";
    private const string TrainingTheilsUStatisticResultName = "Average Theil's U (training)";
    private const string TestTheilsUStatisticResultName = "Average Theil's U (test)";

    public new ITimeSeriesPrognosisModel Model {
      get { return (ITimeSeriesPrognosisModel)base.Model; }
      protected set { base.Model = value; }
    }

    public new ITimeSeriesPrognosisProblemData ProblemData {
      get { return (ITimeSeriesPrognosisProblemData)base.ProblemData; }
      set { base.ProblemData = value; }
    }

    public abstract IEnumerable<IEnumerable<double>> PrognosedTrainingValues { get; }
    public abstract IEnumerable<IEnumerable<double>> PrognosedTestValues { get; }
    public abstract IEnumerable<IEnumerable<IEnumerable<double>>> GetPrognosedValues(IEnumerable<int> rows, int horizon);

    #region Results
    public double[] TrainingMeanSquaredError {
      get { return ((DoubleArray)this[TrainingMeanSquaredErrorResultName].Value).ToArray(); }
      private set { this[TrainingMeanSquaredErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TestMeanSquaredError {
      get { return ((DoubleArray)this[TestMeanSquaredErrorResultName].Value).ToArray(); }
      private set { this[TestMeanSquaredErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingMeanAbsoluteError {
      get { return ((DoubleArray)this[TrainingMeanAbsoluteErrorResultName].Value).ToArray(); }
      private set { this[TrainingMeanAbsoluteErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TestMeanAbsoluteError {
      get { return ((DoubleArray)this[TestMeanAbsoluteErrorResultName].Value).ToArray(); }
      private set { this[TestMeanAbsoluteErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingRSquared {
      get { return ((DoubleArray)this[TrainingSquaredCorrelationResultName].Value).ToArray(); }
      private set { this[TrainingSquaredCorrelationResultName].Value = new DoubleArray(value); }
    }
    public double[] TestRSquared {
      get { return ((DoubleArray)this[TestSquaredCorrelationResultName].Value).ToArray(); }
      private set { this[TestSquaredCorrelationResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingRelativeError {
      get { return ((DoubleArray)this[TrainingRelativeErrorResultName].Value).ToArray(); }
      private set { this[TrainingRelativeErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TestRelativeError {
      get { return ((DoubleArray)this[TestRelativeErrorResultName].Value).ToArray(); }
      private set { this[TestRelativeErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingNormalizedMeanSquaredError {
      get { return ((DoubleArray)this[TrainingNormalizedMeanSquaredErrorResultName].Value).ToArray(); }
      private set { this[TrainingNormalizedMeanSquaredErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TestNormalizedMeanSquaredError {
      get { return ((DoubleArray)this[TestNormalizedMeanSquaredErrorResultName].Value).ToArray(); }
      private set { this[TestNormalizedMeanSquaredErrorResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingDirectionalSymmetry {
      get { return ((DoubleArray)this[TrainingDirectionalSymmetryResultName].Value).ToArray(); }
      private set { this[TrainingDirectionalSymmetryResultName].Value = new DoubleArray(value); }
    }
    public double[] TestDirectionalSymmetry {
      get { return ((DoubleArray)this[TestDirectionalSymmetryResultName].Value).ToArray(); }
      private set { this[TestDirectionalSymmetryResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingWeightedDirectionalSymmetry {
      get { return ((DoubleArray)this[TrainingWeightedDirectionalSymmetryResultName].Value).ToArray(); }
      private set { this[TrainingWeightedDirectionalSymmetryResultName].Value = new DoubleArray(value); }
    }
    public double[] TestWeightedDirectionalSymmetry {
      get { return ((DoubleArray)this[TestWeightedDirectionalSymmetryResultName].Value).ToArray(); }
      private set { this[TestWeightedDirectionalSymmetryResultName].Value = new DoubleArray(value); }
    }
    public double[] TrainingTheilsUStatistic {
      get { return ((DoubleArray)this[TrainingTheilsUStatisticResultName].Value).ToArray(); }
      private set { this[TrainingTheilsUStatisticResultName].Value = new DoubleArray(value); }
    }
    public double[] TestTheilsUStatistic {
      get { return ((DoubleArray)this[TestTheilsUStatisticResultName].Value).ToArray(); }
      private set { this[TestTheilsUStatisticResultName].Value = new DoubleArray(value); }
    }
    #endregion

    [StorableConstructor]
    protected TimeSeriesPrognosisSolutionBase(bool deserializing) : base(deserializing) { }
    protected TimeSeriesPrognosisSolutionBase(TimeSeriesPrognosisSolutionBase original, Cloner cloner)
      : base(original, cloner) {
    }
    protected TimeSeriesPrognosisSolutionBase(ITimeSeriesPrognosisModel model, ITimeSeriesPrognosisProblemData problemData)
      : base(model, problemData) {
      Add(new Result(TrainingMeanSquaredErrorResultName, "Mean of squared errors of the model on the training partition", new DoubleArray()));
      Add(new Result(TestMeanSquaredErrorResultName, "Mean of squared errors of the model on the test partition", new DoubleArray()));
      Add(new Result(TrainingMeanAbsoluteErrorResultName, "Mean of absolute errors of the model on the training partition", new DoubleArray()));
      Add(new Result(TestMeanAbsoluteErrorResultName, "Mean of absolute errors of the model on the test partition", new DoubleArray()));
      Add(new Result(TrainingSquaredCorrelationResultName, "Squared Pearson's correlation coefficient of the model output and the actual values on the training partition", new DoubleArray()));
      Add(new Result(TestSquaredCorrelationResultName, "Squared Pearson's correlation coefficient of the model output and the actual values on the test partition", new DoubleArray()));
      Add(new Result(TrainingRelativeErrorResultName, "Average of the relative errors of the model output and the actual values on the training partition", new DoubleArray()));
      Add(new Result(TestRelativeErrorResultName, "Average of the relative errors of the model output and the actual values on the test partition", new DoubleArray()));
      Add(new Result(TrainingNormalizedMeanSquaredErrorResultName, "Normalized mean of squared errors of the model on the training partition", new DoubleArray()));
      Add(new Result(TestNormalizedMeanSquaredErrorResultName, "Normalized mean of squared errors of the model on the test partition", new DoubleArray()));
      Add(new Result(TrainingDirectionalSymmetryResultName, "The average directional symmetry of the forecasts of the model on the training partition", new DoubleArray()));
      Add(new Result(TestDirectionalSymmetryResultName, "The average directional symmetry of the forecasts of the model on the test partition", new DoubleArray()));
      Add(new Result(TrainingWeightedDirectionalSymmetryResultName, "The average weighted directional symmetry of the forecasts of the model on the training partition", new DoubleArray()));
      Add(new Result(TestWeightedDirectionalSymmetryResultName, "The average weighted directional symmetry of the forecasts of the model on the test partition", new DoubleArray()));
      Add(new Result(TrainingTheilsUStatisticResultName, "The average Theil's U statistic of the forecasts of the model on the training partition", new DoubleArray()));
      Add(new Result(TestTheilsUStatisticResultName, "The average Theil's U statistic of the forecasts of the model on the test partition", new DoubleArray()));
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {

    }

    protected void CalculateResults() {
      OnlineCalculatorError errorState;
      /*
      double[] estimatedTrainingValues = PrognosedTrainingValues.ToArray(); // cache values
      double[] originalTrainingValues = ProblemData.Dataset.GetDoubleValues(ProblemData.TargetVariable, ProblemData.TrainingIndizes).ToArray();
      double[] estimatedTestValues = PrognosedTestValues.ToArray(); // cache values
      double[] originalTestValues = ProblemData.Dataset.GetDoubleValues(ProblemData.TargetVariable, ProblemData.TestIndizes).ToArray();

      double trainingMse = OnlineMeanSquaredErrorCalculator.Calculate(originalTrainingValues, estimatedTrainingValues, out errorState);
      TrainingMeanSquaredError = errorState == OnlineCalculatorError.None ? trainingMse : double.NaN;
      double testMse = OnlineMeanSquaredErrorCalculator.Calculate(originalTestValues, estimatedTestValues, out errorState);
      TestMeanSquaredError = errorState == OnlineCalculatorError.None ? testMse : double.NaN;

      double trainingMae = OnlineMeanAbsoluteErrorCalculator.Calculate(originalTrainingValues, estimatedTrainingValues, out errorState);
      TrainingMeanAbsoluteError = errorState == OnlineCalculatorError.None ? trainingMae : double.NaN;
      double testMae = OnlineMeanAbsoluteErrorCalculator.Calculate(originalTestValues, estimatedTestValues, out errorState);
      TestMeanAbsoluteError = errorState == OnlineCalculatorError.None ? testMae : double.NaN;

      double trainingR2 = OnlinePearsonsRSquaredCalculator.Calculate(originalTrainingValues, estimatedTrainingValues, out errorState);
      TrainingRSquared = errorState == OnlineCalculatorError.None ? trainingR2 : double.NaN;
      double testR2 = OnlinePearsonsRSquaredCalculator.Calculate(originalTestValues, estimatedTestValues, out errorState);
      TestRSquared = errorState == OnlineCalculatorError.None ? testR2 : double.NaN;

      double trainingRelError = OnlineMeanAbsolutePercentageErrorCalculator.Calculate(originalTrainingValues, estimatedTrainingValues, out errorState);
      TrainingRelativeError = errorState == OnlineCalculatorError.None ? trainingRelError : double.NaN;
      double testRelError = OnlineMeanAbsolutePercentageErrorCalculator.Calculate(originalTestValues, estimatedTestValues, out errorState);
      TestRelativeError = errorState == OnlineCalculatorError.None ? testRelError : double.NaN;

      double trainingNmse = OnlineNormalizedMeanSquaredErrorCalculator.Calculate(originalTrainingValues, estimatedTrainingValues, out errorState);
      TrainingNormalizedMeanSquaredError = errorState == OnlineCalculatorError.None ? trainingNmse : double.NaN;
      double testNmse = OnlineNormalizedMeanSquaredErrorCalculator.Calculate(originalTestValues, estimatedTestValues, out errorState);
      TestNormalizedMeanSquaredError = errorState == OnlineCalculatorError.None ? testNmse : double.NaN;
              */

      double[] trainingDs = new double[ProblemData.TargetVariables.Count()];
      double[] testDs = new double[ProblemData.TargetVariables.Count()];

      double[] trainingWds = new double[ProblemData.TargetVariables.Count()];
      double[] testWds = new double[ProblemData.TargetVariables.Count()];

      double[] trainingTheilsU = new double[ProblemData.TargetVariables.Count()];
      double[] testTheilsU = new double[ProblemData.TargetVariables.Count()];
      int i = 0;
      var predictedContinuationTraining = PrognosedTrainingValues.ToArray();
      var predictedContinuationTest = PrognosedTestValues.ToArray();

      foreach (var targetVariable in ProblemData.TargetVariables) {
        var actualTrainingValues = ProblemData.Dataset.GetDoubleValues(targetVariable, ProblemData.TrainingIndizes).ToArray();
        var startTrainingValues = actualTrainingValues.Take(1);
        // only one continuation (but the full training set)
        var actualContinuationTraining = actualTrainingValues.Skip(1);

        var actualTestValues = ProblemData.Dataset.GetDoubleValues(targetVariable, ProblemData.TestIndizes).ToArray();
        var startTestValues = actualTestValues.Take(1);
        // only one continuation (but the full training set)
        var actualContinuationTest = actualTestValues.Skip(1);


        trainingDs[i] = OnlineDirectionalSymmetryCalculator.Calculate(startTrainingValues,
          Enumerable.Repeat(actualContinuationTraining, 1),
          Enumerable.Repeat(predictedContinuationTraining[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) trainingDs[i] = double.NaN;

        testDs[i] = OnlineDirectionalSymmetryCalculator.Calculate(startTestValues,
          Enumerable.Repeat(actualContinuationTest, 1),
          Enumerable.Repeat(predictedContinuationTest[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) testDs[i] = double.NaN;

        trainingWds[i] =
          OnlineWeightedDirectionalSymmetryCalculator.Calculate(startTrainingValues,
          Enumerable.Repeat(actualContinuationTraining, 1),
          Enumerable.Repeat(predictedContinuationTraining[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) trainingWds[i] = double.NaN;

        testWds[i] = OnlineWeightedDirectionalSymmetryCalculator.Calculate(startTestValues,
          Enumerable.Repeat(actualContinuationTest, 1),
          Enumerable.Repeat(predictedContinuationTest[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) testWds[i] = double.NaN;

        trainingTheilsU[i] = OnlineTheilsUStatisticCalculator.Calculate(startTrainingValues,
          Enumerable.Repeat(actualContinuationTraining, 1),
          Enumerable.Repeat(predictedContinuationTraining[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) trainingTheilsU[i] = double.NaN;

        testTheilsU[i] = OnlineTheilsUStatisticCalculator.Calculate(startTestValues,
          Enumerable.Repeat(actualContinuationTest, 1),
          Enumerable.Repeat(predictedContinuationTest[i], 1), out errorState);
        if (errorState != OnlineCalculatorError.None) testTheilsU[i] = double.NaN;
        i++;
      }

      TrainingDirectionalSymmetry = trainingDs;
      TestDirectionalSymmetry = testDs;
      TrainingWeightedDirectionalSymmetry = trainingWds;
      TestWeightedDirectionalSymmetry = testWds;
      TrainingTheilsUStatistic = trainingTheilsU;
      TestTheilsUStatistic = testTheilsU;
    }
  }
}
